/*
EvolveOp.h
Written by Walter O. Krawec

Copyright (c) 2014 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef EVOLVE_OP_H_
#define EVOLVE_OP_H_

#include "Algebra/SimpleAlgebra.h"
#include <list>

#include <fstream>

#ifdef WIN32
#include <conio.h>
#endif

class EvolveOp
{
public:
	// The following data structures are for rule specifications (use EvolveOpHelper to construct):

	struct InputOp
	{
		enum OpType
		{
			StaticOp=0,
			EvolvedOp,
			Measure,
			RuleOp
		};

		InputOp() : type(EvolvedOp), index(0), subIndex(0) {};
		InputOp(OpType _type, int _index) : type(_type), index(_index), subIndex(0) {};
		InputOp(OpType _type, int _index, int _subIndex) : type(_type), index(_index), subIndex(_subIndex) {};

		OpType type;
		int index;
		int subIndex;
	};

	// inputs[i] specifies which operator to apply on iteration i.  Thus, we apply input[0] then input[1], etc. then measure
	// StaticOp uses V list, EvolvedOp uses U list
	struct Inputs
	{
		std::vector < InputOp > inputs;
	};

	struct Action
	{
		Action() : actionSpace(0), actionBasis(0) {};
		Action(int _actionSpace, int _actionBasis) : actionSpace(_actionSpace), actionBasis(_actionBasis) {};

		int actionSpace;
		int actionBasis;
	};

	// maps {actions[0], ..., actions[n]} -> p
	// which means that the prob. of measuring actions[0], OR actions[1], etc. should be "p"
	// each actions[i] is a list of entangled actions; thus this structure specifies:
	// the prob. of measuring (actions[0][0] AND actions[0][1] ...) OR (actions[1][0] AND actions[1][1] ...), etc. should be "p"

	struct Probability
	{
		std::vector < std::vector <Action> > actions;
		double p;
	};

	struct PrDist
	{
		std::vector <Probability> pr;
	};

	// a single rule thus maps inputs -> prob. dist
	// thus a single rule may specify, at most:
	// { {ops}^+ * V * rho }^+ --> rho'
	// where pr. of measuring "a" is specified by "pr"
	// also specifies which sub spaces to measure; if empty, measures all
	// copies specifies how many times to repeat this rule (i.e., apply inputs, measure pr, apply inputs, measure pr, etc.)
	struct SingleRule
	{
		SingleRule() : copies(1) {};
		Inputs inputs;
		PrDist pr;
		int copies;

		std::vector <int> spacesToMeasure;

		void measureUsedSpaces()
		{
			for(unsigned int i=0; i<pr.pr.size(); ++i)
			{
				for(unsigned int j=0; j<pr.pr[i].actions.size(); ++j)
				{
					for(unsigned int k=0; k<pr.pr[i].actions[j].size(); ++k)
					{
						bool found = false;
						for(unsigned int l=0; l<spacesToMeasure.size(); ++l)
						{
							if(pr.pr[i].actions[j][k].actionSpace == spacesToMeasure[l])
							{
								found = true;
								break;
							}
						}
						if(!found)
							spacesToMeasure.push_back(pr.pr[i].actions[j][k].actionSpace);
					}
				}
			}
		}
	};

	// a list of SingleRules.  Make sure to call "finalize()" after pushing all single rules
	// to the "rules" vector; this routine will set "maxIterations" correctly.
	// if initState is not defined, default initState is used (EvolveOp::setup)
	struct Rules
	{
		Rules() {};
		Rules(std::vector <SingleRule>& _r) : rules(_r) {};
		Rules(std::vector <SingleRule>& _r, algebra::mat& _initState) : rules(_r) {initState = _initState;}
		std::vector <SingleRule> rules;
		algebra::mat initState;
	};

	// the following data structures are for the GA:
	struct OpData
	{
		OpData() : systemSize(1) {};

		std::vector <double> phi, psi, chi;
		algebra::mat op, opT;	// opT = op^T
		algebra::mat opSmall, opSmallT;	// same as op just not tensored with I
		int systemSize;

		void randomize(int opSize, int _systemSize)
		{
			systemSize = _systemSize;
			int size1 = opSize*opSize;
			size1 = size1/2;
			phi.resize(size1);
			psi.resize(size1);
			chi.resize(opSize);

			for(int i=0; i<(int)phi.size(); ++i)
			{
				phi[i] = 3.1415*((double)(rand()%10000))/10000.0;
				psi[i] = 3.1415*((double)(rand()%10000))/10000.0;
				if(i < opSize)
					chi[i] = 3.1415*((double)(rand()%10000))/10000.0;
			}

			op.create(opSize);
			//op.random_unitary(phi, psi, chi);
			refreshOp();
		}
		void refreshOp()
		{
			op.random_unitary(phi, psi, chi);
			//op.transpose(opT);

			opSmall = op;
			opSmall.transpose(opSmallT);
			if(systemSize == op.getRows())
			{
				op.transpose(opT);
				return;
			}
			// set U := U x I

			algebra::mat I, temp;
			I.create(systemSize / op.getRows());
			I.ident();
			op.tensor(&I, &temp);
			op = temp;
			op.transpose(opT);
		}
	};

	struct RuleOp
	{
		RuleOp() : T1(NULL), T2(NULL), arg1(0), arg2(0) {};
		~RuleOp() {cleanup();}
		void cleanup()
		{
			if(T1)
				delete T1;
			if(T2)
				delete T2;
			T1 = NULL;
			T2 = NULL;
		}

		enum Operation
		{
			StaticOp=0,	// arg1 = index
			EvolvedOp,	// arg1 = index
			Proj,		// arg1 = space index, arg2 = basis index
			Ident,		// arg1 = size
			Add,		// T1 = op1, T2 = op2
			Subtract,	// ..
			Mult,		// ..
			Tensor		// ..
		};

		RuleOp* T1,*T2;
		int arg1, arg2;
		Operation type;

		void evaluate(std::vector <int>& A, std::vector <OpData>& U, std::vector <algebra::mat>& V, algebra::mat& opOutput, algebra::mat& opOutputT);
	};

	// candidate solution: a list of OpData (which are ops)
	struct CS
	{
		std::vector <OpData> U;
		double fitness, fitness2;
		std::vector <double> ruleFitness;

		bool operator< (const CS& other)
		{
			if(fitness < other.fitness)
				return true;
			return false;
		}

		// startU is only used to lookup operator size
		void setupRandom(std::vector <algebra::mat>& startU, int systemSize)
		{
			int numOps = (int)startU.size();
			U.resize(numOps);
			for(int i=0; i<numOps; ++i)
				U[i].randomize(startU[i].getRows(), systemSize);
			fitness = 0;
		}

		void randomEvolve()
		{
			int p = 20;
			//const double step = 3.1415/24.0;
			int r = rand()%10000;
			double step = r;
			//step = step/10000.0 * 3.1415/12.0*2;
			step = step/10000.0 * 3.1415/10.0;
			double dir;

			//step = 3.1415/5.0;
			for(unsigned int i=0; i<U.size(); ++i)
			{
				for(unsigned int j=0; j<U[i].phi.size(); ++j)
				{
					/*if(rand()%100 < p)
						U[i].phi[j] = 3.1415*(((double)(rand()%10000))/10000.0);
					if(rand()%100 < p)
						U[i].psi[j] = 3.1415*(((double)(rand()%10000))/10000.0);
					if(j < U[i].chi.size() && rand()%100 < p)
						U[i].chi[j] = 3.1415*(((double)(rand()%10000))/10000.0);*/
					dir = 1;
					if(rand()%100 < 50) dir = -1;
					if(rand()%100 < p)
						U[i].phi[j] += dir*step;

					dir = 1;
					if(rand()%100 < 50) dir = -1;
					if(rand()%100 < p)
						U[i].psi[j] += dir*step;

					dir = 1;
					if(rand()%100 < 50) dir = -1;
					if(j < U[i].chi.size() && rand()%100 < p)
						U[i].chi[j] += dir*step;
				}
				U[i].refreshOp();
			}
		}
		void randomEvolve(CS& child)
		{
			child = *this;
			//const double step = 3.1415/24.0;
			int r = rand()%10000;
			double step = r;
			step = step/10000.0 * 3.1415/10.0;
			double dir;
			//step = 3.1415/5.0;
			//child.sensors.resize(sensors.size());
			int p = 80;
			for(unsigned int i=0; i<U.size(); ++i)
			{
				for(unsigned int j=0; j<U[i].phi.size(); ++j)
				{
					/*if(rand()%100 < p)
						child.U[i].phi[j] = 3.1415*(((double)(rand()%10000))/10000.0);
					if(rand()%100 < p)
						child.U[i].psi[j] = 3.1415*(((double)(rand()%10000))/10000.0);
					if(j < U[i].chi.size() && rand()%100 < p)
						child.U[i].chi[j] = 3.1415*(((double)(rand()%10000))/10000.0);
					continue;*/
					//r = rand()%10000;
					//step = r;
					//step = step/10000.0 * 3.1415/10.0;

					dir = 1;
					if(rand()%100 < 50) dir = -1;
					if(rand()%100 < p)
						child.U[i].phi[j] += dir*step;


					//r = rand()%10000;
					//step = r;
					//step = step/10000.0 * 3.1415/10.0;

					dir = 1;
					if(rand()%100 < 50) dir = -1;
					if(rand()%100 < p)
						child.U[i].psi[j] += dir*step;


					//r = rand()%10000;
					//step = r;
					//step = step/10000.0 * 3.1415/10.0;

					dir = 1;
					if(rand()%100 < 50) dir = -1;
					if(j < child.U[i].chi.size() && rand()%100 < p)
						child.U[i].chi[j] += dir*step;
				}
				child.U[i].refreshOp();
			}
		}

		void crossOver(CS& other, CS& child)
		{
			child = *this;
			int type = rand()%100;
			int r2 = rand()%10000;
			double r = r2;
			r = r / 10000.0;
			bool blendType = false;
			for(unsigned int i=0; i<U.size(); ++i)
			{
				int bound1 = rand()%U[i].phi.size();
				int bound2 = rand()%U[i].psi.size();
				for(int j=0; j<(int)U[i].phi.size(); ++j)
				{
					if(blendType)
					{
						child.U[i].phi[j] = r*U[i].phi[j] + (1-r)*other.U[i].phi[j];
						child.U[i].psi[j] = r*U[i].psi[j] + (1-r)*other.U[i].psi[j];
						continue;
					}
					if(j < bound1)
						child.U[i].phi[j] = U[i].phi[j];
					else
						child.U[i].phi[j] = other.U[i].phi[j];

					if(j < bound2)
						child.U[i].psi[j] = U[i].psi[j];
					else
						child.U[i].psi[j] = other.U[i].psi[j];

					//if(j < sensors[i].data.phi.size()/2)
					/*if(j < bound1)
					{
						child.U[i].phi[j] = U[i].phi[j];
						child.U[i].psi[j] = U[i].psi[j];
					}
					else
					{
						child.U[i].phi[j] = other.U[i].phi[j];
						child.U[i].psi[j] = other.U[i].psi[j];
					}*/
				}

				int bound = rand()%U[i].chi.size();
				for(int j=0; j<(int)U[i].chi.size(); ++j)
				{
					if(blendType)
					{
						child.U[i].chi[j] = r*U[i].chi[j] + (1-r)*other.U[i].chi[j];
						continue;
					}
					if(j < bound)
						child.U[i].chi[j] = U[i].chi[j];
					else
						child.U[i].chi[j] = other.U[i].chi[j];
				}
				child.U[i].refreshOp();
			}
		}
	};
public:
	EvolveOp() {};

	bool setup(std::vector <algebra::mat>* _V, std::vector <RuleOp*>* _T, algebra::mat& _initState, std::vector <int>& _actionDim, std::vector <int>* _resetAction, int numOps)
	{
		int dim = 1;
		for(unsigned int i=0; i<_actionDim.size(); ++i)
			dim *= _actionDim[i];

		std::vector <int> opSizes;

		for(int i=0; i<numOps; ++i)
			opSizes.push_back(dim);

		return setup(_V, _T, _initState, _actionDim, _resetAction, opSizes);
	}
	// opSizes should be a list s.t. opSizes.size() equals the amount of operators to evolve and opSizes[i] is the dimension of the i'th op (between 1 and systemSize)
	bool setup(std::vector <algebra::mat>* _V, std::vector <RuleOp*>* _T, algebra::mat& _initState, std::vector <int>& _actionDim, std::vector <int>* _resetAction, std::vector <int>& opSizes)
	{
		if(!_V)
			V.clear();
		else
			V = *_V;

		if(!_T)
			T.clear();
		else
			T = *_T;

		for(unsigned int i=0; i<V.size(); ++i)
		{
			algebra::mat matVT;
			V[i].transpose(matVT);
			VT.push_back(matVT);
		}
		//V.transpose(VT);
		initState = _initState;
		//memoryDim = _memoryDim;
		actionDim = _actionDim;

		if(!_resetAction)
		{
			resetAction.clear();
			for(unsigned int i=0; i<actionDim.size(); ++i)
				resetAction.push_back(-1);
		}
		else
			resetAction = *_resetAction;

		systemSize = 1;
		for(unsigned int i=0; i<actionDim.size(); ++i)
			systemSize *= actionDim[i];

		// set initState = |initState><initState|
		algebra::mat temp, temp2;
		initState.transpose(temp);
		initState.multiply(temp, temp2);
		initState = temp2;

		//U.resize(numOps);
		for(unsigned int i=0; i<opSizes.size(); ++i)
		{
			temp.create(opSizes[i]);
			U.push_back(temp);
		}

		//if(V.getRows() != memoryDim*actionDim || initState.getRows() != memoryDim*actionDim) return false;
		//if(V.getRows() != V.getCols()) return false;

		setupMeasurementOps();
		setupResetOps();
		/*if(resetAction == -1)
			setupResetOps(false, 0);
		else
			setupResetOps(true, resetAction);*/

		return true;
	}

	// given a list of actions, computes the pr of measuring those states from quantum state rho.
	// actions should be a list of actions across different action spaces (e.g., <0, a>, <2, b>, etc. where <i, j> is action space "i", action basis "j")
	// do not, I beg, call this to measure probability of different action bases in the same action space (e.g., do not set actions to be <0, a>, <0, b>)
	// for such application you should sum this function: e.g. calculatePr(<0,a>) + calculatePr(<0,b>) which computes correction the Pr of choosing |a> or |b>
	// (one cannot choose both |a> and |b>, but one can chooe both |a,*,b> as in the first example)
	double calculatePr(algebra::mat& rho, std::vector <Action>& actions);

	void prepareGA(int startPopulation)
	{
		//startPopulation = 1;
		population.clear();
		for(int i=0; i<startPopulation; ++i)
		{
			CS cs;
			cs.setupRandom(U, systemSize);
			population.push_back(cs);
		}
	}

	void evolve(std::vector <Rules>& rules, int maxIterations, int maxPopulation, std::vector <double>* fitnessPoints = NULL, std::vector <double>* avgFitnessPoints=NULL, std::vector <int>* fitnessAvgNum = NULL);

	bool saveAsQProg(const std::string& filename);
	bool saveAsSpin(const std::string& filename);
	bool saveAsText(const std::string& filename);

	void test(std::vector <Rules>& rules);
//private:
	void setupMeasurementOps();
	void setupResetOps();//(bool resetAction, int actionBasis);
	void calculateFitness(CS& cs, std::vector <Rules>& rules, bool debug=false);
	// sets rho to be op*rho*op^T
	void applyOp(algebra::mat& rho, algebra::mat& op, algebra::mat& opT)
	{
		//std::cout << "applyOp, before:\n";
		//rho.print(std::cout);
		op.multiply(rho, applyOpTempMat);
		applyOpTempMat.multiply(opT, rho);
		//std::cout << "after:\n";
		//rho.print(std::cout);
	}

	// returns mixed state resulting from measurement of specified action spaces
	void measure(algebra::mat& rho, std::vector <int>& spacesToMeasure);
	void measure(algebra::mat& rho, int actionSpace);

	// returns mixed state resulting from applying specified measurement op
	// i.e., rho := M*rho / (tr(M*rho))
	void measure(algebra::mat& rho, algebra::mat& M, algebra::mat& R, algebra::mat& RT);

	//algebra::mat V, VT;		// model-specific op; e.g., TOP.  VT = V^T
	std::vector <algebra::mat> V, VT;
	std::vector <RuleOp*> T;
	algebra::mat initState;
	algebra::mat applyOpTempMat; // used by applyOp
	std::vector <algebra::mat> U;	// evolved ops
	std::vector < std::vector <algebra::mat> > M;	// measurement ops; M[i][j] = I x ... x I x |j><j| x I ... x I where the op |j><j| appears in the i'th subspace
	std::vector < std::vector <algebra::mat> > R, RT;	// reset ops; R[i][j] = I x ... x I x |r><j| x I ... x I; RT[j] = R[j]^T

	std::list <CS> population;

	//int memoryDim, actionDim;
	std::vector <int> actionDim;
	std::vector <int> resetAction;
	int systemSize;
	//int resetAction;	// if >= 0, H_A is reset to |resetAction> after each measurement
};

#endif