/*
EvolveOpHelper.cpp
Written by Walter O. Krawec

Copyright (c) 2014 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "EvolveOpHelper.h"

bool EvolveOpHelper::parseRuleOpString(const std::string& rules, std::vector <EvolveOp::RuleOp*>& ruleOpOutput)
{
	wok::Text text;
	setupTextParser(text);

	text.parse(rules);
	for(unsigned int i=0; i<text.size(); ++i)
	{
		EvolveOp::RuleOp *T;
		T = parseRuleOp(text[i]);
		ruleOpOutput.push_back(T);
	}

	return true;
}

// T := U.i | V.i | P.i.j | I.n | *, T, T | +, T, T | -, T, T | X, T, T
EvolveOp::RuleOp* EvolveOpHelper::parseRuleOp(const std::string& rule)
{
	wok::Text text;
	setupTextParser(text);
	text.addDelimiter('.');

	text.parse(rule);

	if(text[0] == "U")
	{
		EvolveOp::RuleOp *T = new EvolveOp::RuleOp;
		T->type = EvolveOp::RuleOp::EvolvedOp;
		T->arg1 = atoi(text[1].c_str());
		return T;
	}
	else if(text[0] == "V")
	{
		EvolveOp::RuleOp *T = new EvolveOp::RuleOp;
		T->type = EvolveOp::RuleOp::StaticOp;
		T->arg1 = atoi(text[1].c_str());
		return T;
	}
	else if(text[0] == "P")
	{
		EvolveOp::RuleOp *T = new EvolveOp::RuleOp;
		T->type = EvolveOp::RuleOp::Proj;
		T->arg1 = atoi(text[1].c_str());
		T->arg2 = atoi(text[2].c_str());
		return T;
	}
	else if(text[0] == "I")
	{
		EvolveOp::RuleOp *T = new EvolveOp::RuleOp;
		T->type = EvolveOp::RuleOp::Ident;
		T->arg1 = atoi(text[1].c_str());
		return T;
	}
	else if(text[0] == "+")
	{
		EvolveOp::RuleOp *T = new EvolveOp::RuleOp;
		T->type = EvolveOp::RuleOp::Add;
		T->T1 = parseRuleOp(text[1]);
		T->T2 = parseRuleOp(text[2]);
		return T;
	}
	else if(text[0] == "-")
	{
		EvolveOp::RuleOp *T = new EvolveOp::RuleOp;
		T->type = EvolveOp::RuleOp::Subtract;
		T->T1 = parseRuleOp(text[1]);
		T->T2 = parseRuleOp(text[2]);
		return T;
	}
	else if(text[0] == "*")
	{
		EvolveOp::RuleOp *T = new EvolveOp::RuleOp;
		T->type = EvolveOp::RuleOp::Mult;
		T->T1 = parseRuleOp(text[1]);
		T->T2 = parseRuleOp(text[2]);
		return T;
	}
	else if(text[0] == "X")
	{
		EvolveOp::RuleOp *T = new EvolveOp::RuleOp;
		T->type = EvolveOp::RuleOp::Tensor;
		T->T1 = parseRuleOp(text[1]);
		T->T2 = parseRuleOp(text[2]);
		return T;
	}

	return NULL;
}

// RuleList := (Rules), (Rules), ...
bool EvolveOpHelper::parseRuleString(const std::string& rules, std::vector <EvolveOp::Rules>& ruleListOutput)
{
	wok::Text text;
	setupTextParser(text);

	text.parse(rules);
	for(unsigned int i=0; i<text.size(); ++i)
	{
		EvolveOp::Rules r;
		if(!parseRules(text[i], r))
			return false;
		ruleListOutput.push_back(r);
	}
	return true;
}

// Rules := (initState), {Rule}^+
// Rule := (copies), (opList), (pr), (measure spaces)
// measure spaces := all | used | integer list
bool EvolveOpHelper::parseRules(const std::string& rule, EvolveOp::Rules& ruleOutput)
{
	wok::Text text, text2;
	setupTextParser(text);
	setupTextParser(text2);
	text.parse(rule);

	if(!parseInitState(text[0], ruleOutput.initState)) return false;

	for(unsigned int i=1; i<text.size(); i+=4)
	{
		EvolveOp::SingleRule r;
		r.copies = atoi(text[i].c_str());
		if(!parseOpList(text[i+1], r))
			return false;
		if(!parsePr(text[i+2], r))
			return false;

		text2.parse(text[i+3]);
		if(text2[0] == "all")
			r.spacesToMeasure.clear();
		else if(text2[0] == "used")
			r.measureUsedSpaces();
		else
		{
			// ToDo: integer list
		}
		ruleOutput.rules.push_back(r);
	}

	return true;
}

// Pr := p1:(action_list), p2:(action_list), ...
// action_list := (space:index, space:index, ...), (space:index, ...), ...
bool EvolveOpHelper::parsePr(const std::string& rule, EvolveOp::SingleRule& ruleOutput)
{
	wok::Text text, text2, text3;
	setupTextParser(text);
	setupTextParser(text2);
	setupTextParser(text3);
	text.parse(rule);

	for(unsigned int i=0; i<text.size(); i+=2)
	{
		EvolveOp::Probability p;
		p.p = atof(text[i].c_str());
		text2.parse(text[i+1]);			// action_list of the form (), (), ()...
		for(unsigned int j=0; j<text2.size(); ++j)
		{
			text3.parse(text2[j]);	// parse entangled action list of the form space:index, space:index, ...
			std::vector <EvolveOp::Action> actions;
			for(unsigned int k=0; k<text3.size(); k+=2)
			{
				EvolveOp::Action a;
				a.actionSpace = atoi(text3[k].c_str());
				a.actionBasis = atoi(text3[k+1].c_str());
				actions.push_back(a);
			}
			p.actions.push_back(actions);
		}
		ruleOutput.pr.pr.push_back(p);
	}

	return true;
}

// OpList := U/V/M.i.j,U/V/M.i.j,...
bool EvolveOpHelper::parseOpList(const std::string& rule, EvolveOp::SingleRule& ruleOutput)
{
	wok::Text text;
	setupTextParser(text);
	text.addDelimiter('.');
	text.parse(rule);

	for(unsigned int i=0; i<text.size(); ++i)
	{
		if(text[i].size() == 0) return false;
		if(text[i][0] == 'U')
		{
			++i;
			EvolveOp::InputOp op;
			op.type = EvolveOp::InputOp::EvolvedOp;
			op.index = atoi(text[i].c_str());
			ruleOutput.inputs.inputs.push_back(op);
		}
		else if(text[i][0] == 'V')
		{
			++i;
			EvolveOp::InputOp op;
			op.type = EvolveOp::InputOp::StaticOp;
			op.index = atoi(text[i].c_str());
			ruleOutput.inputs.inputs.push_back(op);
		}
		else if(text[i][0] == 'T')
		{
			++i;
			EvolveOp::InputOp op;
			op.type = EvolveOp::InputOp::RuleOp;
			op.index = atoi(text[i].c_str());
			ruleOutput.inputs.inputs.push_back(op);
		}
		else if(text[i][0] == 'M')
		{
			++i;
			EvolveOp::InputOp op;
			op.type = EvolveOp::InputOp::Measure;
			op.index = atoi(text[i].c_str());
			++i;
			if(text[i] == "*")
				op.subIndex = -1;
			else
				op.subIndex = atoi(text[i].c_str());
			ruleOutput.inputs.inputs.push_back(op);
		}
	}

	return true;
}

//initState := psi_0 | matrixFile, "FILENAME" | QProg, "FILENAME" | matrix <MATRIX> | pure, real:imag, real:imag, ... | mixed, p1:(initState), p2:(initState), ...)
bool EvolveOpHelper::parseInitState(const std::string& rule, algebra::mat& initStateOutput)
{
	wok::Text text;
	setupTextParser(text);
	text.parse(rule);

	std::string str(text[0]);

	if(text[0] == "psi_0") return true;	// use default initState
	else if(text[0] == "pure")
	{
		std::vector <algebra::Complex16> v;
		for(unsigned int i=1; i<text.size(); i+=2)
		{
			double re = atof(text[i].c_str());
			double im = atof(text[i+1].c_str());
			v.push_back(algebra::Complex16(re,im));
		}

		algebra::mat temp, temp2;
		temp.create((int)v.size(), 1);
		for(unsigned int i=0; i<v.size(); ++i)
			temp(i,0) = v[i];
		temp.transpose(temp2);
		temp.multiply(temp2, initStateOutput);
		return true;
	}

	return false;
}